-- TODO: Item reordering
-- TODO: Split code in more files
-- TODO: Level Tracking/Skill Tracking (seems impossible without tracking profession pages)
-- TODO: Add more PvE stuff
-- TODO: Use GetSellValue API for money

Wishlist = LibStub('AceAddon-3.0'):NewAddon('Wishlist', 'AceConsole-3.0', 'AceEvent-3.0')
WishlistData = WishlistData or { Factions = {}, Predefined = {} }

local W = Wishlist
local data = WishlistData

local L = LibStub("AceLocale-3.0"):GetLocale("Wishlist", true)
local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local acd = LibStub("AceConfigDialog-3.0")
local BF = LibStub('LibBabble-Faction-3.0'):GetLookupTable()

-- TitanPanel Options
W.id = 'Wishlist'
W.addon = 'Wishlist'
W.button_label = W.id..": "
W.menu_text = W.id
W.artwork_path = 'Interface/PvPRankBadges/PvPRank' .. UnitFactionGroup('player')
W.version = tostring(GetAddOnMetadata(W.addon, "Version")) or "Unknown" 
W.author = GetAddOnMetadata(W.addon, "Author") or "Unknown"

----------------------
--- Ace3 Handlers ---
----------------------

function W:OnInitialize()
	-- Money formatting
	local abacus = LibStub:GetLibrary('LibAbacus-3.0', true)
	self.FormatMoney = abacus.FormatMoneyCondensed

	self.ShortNames = {
		[L['Warsong Gulch Mark of Honor']] = L['WSG Mark(s)'],
		[L['Arathi Basin Mark of Honor']] = L['AB Mark(s)'],
		[L['Alterac Valley Mark of Honor']] = L['AV Mark(s)'],
		[L['Eye of the Storm Mark of Honor']] = L['EotS Mark(s)'],
		[L['Strand of the Ancients Mark of Honor']] = L['SotA Mark(s)'],
	}

	self.Textures = {
		Honor = 'Interface/PvPRankBadges/PvPRank' .. UnitFactionGroup('player'),
		Money = 'Interface/MoneyFrame/UI-MoneyIcons', -- TODO: Better icon
		ArenaPoints = 'Interface/PVPFrame/PVP-ArenaPoints-Icon',
		[L['Warsong Gulch Mark of Honor']] = 'Interface/Icons/INV_Misc_Rune_07',
		[L['Arathi Basin Mark of Honor']] = 'Interface/Icons/INV_Jewelry_Amulet_07',
		[L['Alterac Valley Mark of Honor']] = 'Interface/Icons/INV_Jewelry_Necklace_21',
		[L['Eye of the Storm Mark of Honor']] = 'Interface/Icons/Spell_Nature_EyeOfTheStorm',
		[L['Strand of the Ancients Mark of Honor']] = 'Interface/Icons/INV_Jewelry_Amulet_01',
		[L["Stone Keeper's Shard"]] = 'Interface/Icons/INV_Misc_Platnumdisks',
		[L['Wintergrasp Mark of Honor']] = 'Interface/Icons/INV_Jewelry_Ring_66',
		[L['Badge of Justice']] = 'Interface/Icons/Spell_Holy_ChampionsBond',
		[L['Halaa Battle Token']] = 'Interface/Icons/INV_Misc_Rune_08',
		[L['Halaa Research Token']] = 'Interface/Icons/INV_Misc_Rune_09',
		[L['Emblem of Heroism']] = 'Interface/Icons/Spell_Holy_ProclaimChampion',
		[L['Emblem of Valor']] = 'Interface/Icons/Spell_Holy_ProclaimChampion_02',
		[L['Emblem of Triumph']] = 'Interface/Icons/Spell_Holy_summonChampion',
		[L['Emblem of Conquest']] = 'Interface/Icons/Spell_Holy_ChampionsGrace',
	}
	
	-- Cache standing text
	self.StandingText = {}
	local s
	for s = 1, 8 do
		self.StandingText[s] = self:ColorToHex(FACTION_BAR_COLORS[s]) .. getglobal('FACTION_STANDING_LABEL' .. s)
	end
	
	self.Factions = { BF['Cenarion Expedition'], BF['Lower City'], BF["The Sha'tar"], BF['Keepers of Time'], BF['The Aldor'], BF['The Scryers'], BF['Ashtongue Deathsworn'], BF["Sha'tari Skyguard"], BF['Shattered Sun Offensive'], BF['Argent Crusade'], BF['Frenzyheart Tribe'], BF['Kirin Tor'], BF['Knights of the Ebon Blade'], BF["The Kalu'ak"], BF['The Oracles'], BF['The Sons of Hodir'], BF['The Wyrmrest Accord'] }
	if UnitFactionGroup('player') == 'Horde' then
		-- Battleground Factions
		table.insert(self.Factions, BF['Frostwolf Clan'])
		table.insert(self.Factions, BF['The Defilers'])
		table.insert(self.Factions, BF['Warsong Outriders'])

		-- BC Factions
		table.insert(self.Factions, BF['Thrallmar'])
		table.insert(self.Factions, BF["The Mag'har"])

		-- WotLK Factions
		table.insert(self.Factions, BF['Horde Expedition'])
	else
		-- Battleground Factions
		table.insert(self.Factions, BF['Silverwing Sentinels'])
		table.insert(self.Factions, BF['Stormpike Guard'])
		table.insert(self.Factions, BF['The League of Arathor'])

		-- BC Factions
		table.insert(self.Factions, BF['Honor Hold'])
		table.insert(self.Factions, BF['Kurenai'])

		-- WotLK Factions
		table.insert(self.Factions, BF['Alliance Vanguard'])
	end

	-- Sort all factions
	table.sort(self.Factions)
end

function W:OnEnable()
	-- Create data variables
	self.Data = {
		Factions = {},
		Honor = 0, -- TODO: GetHonorCurrency(),
		ArenaPoints = 0,  -- TODO: GetArenaCurrency(),
		-- TODO: Arena Rating
		Items = {
			-- TODO: Move older items in sub folder
			[L['Warsong Gulch Mark of Honor']] = 0,
			[L['Arathi Basin Mark of Honor']] = 0,
			[L['Alterac Valley Mark of Honor']] = 0,
			[L['Eye of the Storm Mark of Honor']] = 0,
			[L['Badge of Justice']] = 0,
			[L['Halaa Battle Token']] = 0,
			[L['Halaa Research Token']] = 0,
			[L['Strand of the Ancients Mark of Honor']] = 0,
			[L['Wintergrasp Mark of Honor']] = 0,
			[L['Emblem of Heroism']] = 0,
			[L['Emblem of Valor']] = 0,
			[L['Emblem of Triumph']] = 0,
			[L['Emblem of Conquest']] = 0,
			[L['Emblem of Frost']] = 0,
			[L["Stone Keeper's Shard"]] = 0,
		},
		Money = GetMoney(),
	}

	-- Create database
    self.db = LibStub("AceDB-3.0"):New("WishlistDB", {
        profile = {
            Settings = {
                ShowMoneyChanges = false,
                ShowHonorChanges = false,
                ShowItemsChanges = false,
                ShowNeeded = true,
                ShowAll = false,
                UseAtlasLoot = true,
            },
            RequirementItems = {},
            Text = {
                ShowMoney = true,
            },
        },
        char = {
            Items = {},
            Sets = {},
        }
    }, true)

	self:UpdateItems()

	-- Keep reputation up-to-date
	self:RegisterEvent("UPDATE_FACTION")

	-- Keep money up-to-date
	self:RegisterEvent("PLAYER_MONEY")

	self:RegisterEvent("BAG_UPDATE")

	self:UpdateTotals()
	self:Update()

	self:RegisterMenu()
end

function W:OnDisable()
	self:UnregisterAllEvents()
end

function W:Update()
    self:OnDataUpdate()
    self:OnTextUpdate()
    self:OnTooltipUpdate()
end

function W:OnDataUpdate()
	self.Result, self.Failed = self:CheckItem(self.Totals)

	-- Reset results
	self.ItemResults = {}
	
	self:CheckItems(self.ItemResults, self.db.char.Items)
	local _, set
	for _, set in ipairs(self.db.char.Sets) do
		self:CheckItems(self.ItemResults, set.Items, set)
	end
end

function W:OnTextUpdate()
	local text = ''
	if self.db.profile.Text.ShowMoney then
		text = text .. self:GetMoneyText()
	end
	-- TODO: self:SetText(text)
end

function W:OnTooltipUpdate()
    -- TODO: Alternative for tablet
    return

	local cat
	if self.Result then
		cat = tablet:AddCategory('columns', 1, 'showWithoutChildren', true)
		if not self:HasItems() then
			-- Got nothing
			cat:AddLine({ text = L['|cffffffffUse /wishlist add [item] to track items'], justify = 'CENTER' })
		else
			-- Got everything!!
			self:AddItemsToTooltip(self.db.char.Items, cat)
			local index, set
			for index, set in ipairs(self.db.char.Sets) do
				cat:AddLine({ text = '|cffffffff' .. set.Name, justify = 'CENTER' })
				self:AddItemsToTooltip(set.Items, cat)
			end
			cat:AddLine({ text = L['|cffffffffGot everything!'], justify = 'CENTER' })
		end
	else
		-- Something is missing, show totals
		cat = tablet:AddCategory('columns', 3, 'showWithoutChildren', true)
		cat:AddLine({
			text = '', justify = 'CENTER',
			text2 = self.db.profile.Settings.ShowNeeded and L['Need'] or L['Got'], justify2 = 'CENTER',
			text3 = L['Done'],	justify3 = 'CENTER',
		})

		-- Show totals of requirements
		local index, req
		local total = 0
		for index, req in ipairs(self.Failed) do
			local txt, txt2, done = self:RequirementToTooltip(req)
			if done > 1 then
				done = 1
			end
			cat:AddLine({ text = txt, text2 = txt2, text3 = self:PercentageToColoredText(done) })
			total = total + done
		end
		total = total / #self.Failed

		cat:AddLine({ text = L['|cffffffffTotal'], justify = 'CENTER', text2 = '', text3 = self:PercentageToColoredText(total) })

		-- Show percentage for each item
		cat = tablet:AddCategory('text', L['Options Items'], 'columns', 2)
		local index, item
		for index, item in ipairs(self.ItemResults) do
			local name, link = GetItemInfo(item.ItemId)
			local text, set = link or L['Not Found'], item.Set
			if set then
				text = text .. ' |cffcccccc(' .. set .. ')'
			end
			
			local done
			local func
			if item.Result then
				done = self:PercentageToColoredText(1)
				func = function() self:Print((L['%s, you got everything!']):format(link)) end
			else
				local total = 0
				local count = item.Count
				local need = ''
				for index, req in ipairs(item.Failed) do
					local _, _, done = self:RequirementToTooltip(req)
					if done > 1 then
						done = 1
					end
					if strlen(need) > 0 then
					  need = need .. ', '
					end
					need = need .. self:RequirementToNeeded(req)
					total = total + done
				end
				total = (total + count - #item.Failed) / count
				if total == 1 then
				  total = 0.99 -- Just to never show 100%
				end
				done = self:PercentageToColoredText(total)
				func = function() self:Print((L['%s, you need %s.']):format(link, need)) end
			end
			
			cat:AddLine({ text = text, text2 = done, func = func })
		end
	end
end

------------------------
--- Validate Methods ---
------------------------

function W:CheckItem(item, onlyNeeded)
	local result = true
	local failed = {}

	local index, req
	for index, req in ipairs(item.Requirements) do
		if req.Type == 'Note' then
			table.insert(failed, req)
		elseif not self:ValidateRequirement(req, onlyNeeded) then
			table.insert(failed, req)
			result = false
		end
	end

	return result, failed
end

function W:ValidateRequirement(req, onlyNeeded)
	if req and (not self.db.profile.Settings.ShowAll or onlyNeeded) then
		if req.Type == 'Honor' then
			-- Check if we have enough honor
			return self.Data.Honor >= req.Honor
		elseif req.Type == 'Money' then
			-- Check if we have enough money
			return self.Data.Money >= req.Money
		elseif req.Type == 'Arena' then
			-- Check if we have enough arena points
			return self.Data.ArenaPoints >= req.ArenaPoints
		elseif req.Type == 'Faction' then
			-- Check if we have enough reputation
			return self:GetFactionStanding(req.Faction) >= req.Standing
		elseif req.Type == 'Item' then
			-- Check if we have enough of the needed item
			return self:GetItemCount(req.Name) >= req.Required
		end
	end
	return false
end

----------------------
--- Helper Methods ---
----------------------

function W:AddItem(item, items)
    if not item then return end
    
	local name, link = GetItemInfo(item)
	if not name then return end
	local _, _, item = string.find(link, "item:(%-?%d+):")
	item = tonumber(item)

	local requirements = {}

	-- Check if Vendor window is open with item
	local vendorItemCount = GetMerchantNumItems()
	if vendorItemCount > 0 then
		local index
		for index = 1, vendorItemCount do
			local vendorItemLink = GetMerchantItemLink(index)
			if link == vendorItemLink then
				-- Found item in Vendor window, get requirements
				local _, _, price, _, _, _, extendedCost = GetMerchantItemInfo(index)
				if price > 0 then
					requirements = { { Type = 'Money', Money = price } }
				elseif extendedCost == 1 then
					-- Find honor/arena/items required
					local honorPoints, arenaPoints, itemCount = GetMerchantItemCostInfo(index)
					if honorPoints > 0 then
						requirements = { { Type = 'Honor', Honor = honorPoints } }
					elseif arenaPoints > 0 then
						requirements = { { Type = 'Arena', ArenaPoints = arenaPoints } }
					else
						local itemTexture, itemValue, itemLink = GetMerchantItemCostItem(index, 1)
						local itemName, _ = GetItemInfo(itemLink)
						requirements = { { Type = 'Item', Name = itemName, Required = itemValue } }
						self.Textures[itemName] = itemTexture
					end
				end
			end
		end
	end
	
	-- Add item to character's watched items
	self:AddNewItem({ ItemId = item, Requirements = requirements }, link, items)
end

function W:AddRequirementItem(item)
	local name, link = GetItemInfo(item)
	if not name then return end
	local _, _, item = string.find(link, "item:(%-?%d+):")
	item = tonumber(item)

	-- Add requirement
	self.db.profile.RequirementItems[name] = { ItemId = item, Count = GetItemCount(name, true) }
	self:UpdateRequirementsMenu()
	self:UpdateItemsMenu()

	self:Print((L['Added %s to requirements list']):format(link))
end	

function W:AddItemsToTooltip(items, cat)
	local index, item
	for index, item in ipairs(items) do
		local _, link = GetItemInfo(item.ItemId)
		link = link or L['Not Found']
		cat:AddLine({ text = link, justify = 'CENTER', func = function() self:Print((L['%s, you got everything!']):format(link)) end })
	end
end

function W:AddNewItem(item, link, items)
	items = items or self.db.char.Items
	table.insert(items, item)

	self:UpdateTotals()
	self:Update()
	self:UpdateItemsMenu()

	self:Print((L['Added %s to watch list']):format(link))
end

function W:AddRequirements(item, requirements)
	-- Combine new requirements with current ones
	self:CombineRequirements(item.Requirements, requirements)

	self:UpdateTotals()
	self:Update()
	self:UpdateItemsMenu()
end

function W:AddSet(name)
	table.insert(self.db.char.Sets, { Name = name, Items = {} })
	
	self:UpdateItemsMenu()
end

function W:CheckItems(results, items, set)
	local _, item
	for _, item in ipairs(items) do
		local result, failed = self:CheckItem(item, true)
		local itemResult = { ItemId = item.ItemId, Result = result, Failed = failed, Count = #item.Requirements }
		if set then
			itemResult.Set = set.Name
		end
		table.insert(results, itemResult)
	end
end

function W:CheckForItems(items)
	local isChanged = false
	
	local hasItems = {}
	local index
	for index, item in ipairs(items) do
		if GetItemCount(item.ItemId) > 0 then
			local _, itemLink = GetItemInfo(item.ItemId)
			self:Print((L["You got item %s, congratulations!"]):format(itemLink))

			table.insert(hasItems, 1, index)
			isChanged = true
		end
	end
	while #hasItems > 0 do self:RemoveItem(items, table.remove(hasItems), true) end
	
	return isChanged
end

function W:ColorToHex(color)
	return ("|cff%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end

function W:CombineRequirements(original, extra)
	local index, req
	for index, req in ipairs(extra) do
		local existing = self:FindMatchingRequirement(original, req)
		if existing then
			if existing.Type == 'Honor' then
				existing.Honor = existing.Honor + req.Honor
			elseif existing.Type == 'Money' then
				existing.Money = existing.Money + req.Money
			elseif existing.Type == 'Arena' then
				existing.ArenaPoints = existing.ArenaPoints + req.ArenaPoints
			elseif existing.Type == 'Item' then
				existing.Required = existing.Required + req.Required
			elseif existing.Type == 'Faction' then
				existing.Standing = max(existing.Standing, req.Standing)
			end
		else
			-- Copy requirement
			local new = {}
			local key, value
			for key, value in pairs(req) do
				new[key] = value
			end
			table.insert(original, new)
		end
	end
end

function W:FindMatchingRequirement(requirements, match)
	local _, req
	for _, req in ipairs(requirements) do
		if (req.Type == match.Type) and ((req.Type ~= 'Item' and req.Type ~= 'Faction') or (req.Type == 'Item' and req.Name == match.Name) or (req.Type == 'Faction' and req.Faction == match.Faction)) then
			return req
		end
	end
	return nil -- Not found
end

function W:GetFaction(name)
	return self.Data.Factions[name] or { Standing = 4, EarnedValue = 0 } -- Default Neutral
end

-- 0 - Unknown
-- 1 - Hated
-- 2 - Hostile
-- 3 - Unfriendly
-- 4 - Neutral
-- 5 - Friendly
-- 6 - Honored
-- 7 - Revered
-- 8 - Exalted
function W:GetFactionStanding(name)
	local faction = self:GetFaction(name)
	return faction and faction.Standing or 0
end

function W:GetItemCount(name)
	local count = self.Data.Items[name]
	if not count then
		local req = self.db.profile.RequirementItems[name]
		if req then
			count = req.Count
		end
	end
	return count or 0
end

function W:GetMoneyText()
	return self:FormatMoney(self.Data.Money, true)
end

function W:GetShortName(txt)
	return self.ShortNames[txt] or txt
end

function W:GetTexture(name)
	local texture = self.Textures[name]
	if not texture then
		local _, _, _, _, _, _, _, _, _, itemTexture = GetItemInfo(name)
		self.Textures[name] = itemTexture -- Cache for next time
		texture = itemTexture
	end
	return texture
end

function W:HasItems()
	if #self.db.char.Items > 0 then
		return true
	end
	
	local index, set
	for index, set in ipairs(self.db.char.Sets) do
		if #set.Items > 0 then
			return true
		end
	end
	
	return false
end

function W:MoveItem(items, index, newItems)
	table.insert(newItems, table.remove(items, index))
	
	self:Update()
	self:UpdateItemsMenu()
end

-- Converts a percentage (0.00-1.00) to (red-green)
function W:PercentageToColoredText(value)
  local h, s, v = (value / 2.55) * 360, 1, 1

  local sectorPos = h / 60
  local sectorNumber = floor(sectorPos)
  local fractionalSector = sectorPos - sectorNumber

  local p = 0
  local q = 1 - fractionalSector
  local t = 1 - (1 - fractionalSector)

  local r, g, b
  if sectorNumber == 0 then
    r, g, b = v, t, p
  elseif sectorNumber == 1 then
    r, g, b = q, v, p
  elseif sectorNumber == 2 then
    r, g, b = p, v, t
  elseif sectorNumber == 3 then
    r, g, b = p, q, v
  elseif sectorNumber == 4 then
    r, g, b = t, p, v
  else
    r, g, b = v, p, q
  end

  return self:ColorToHex({ r = r, g = g, b = b }) .. ('%d'):format(value * 100) .. '%'
end

function W:PrintFaction(name)
	local standing = self:GetFactionStanding(name)

	if not standing then
		self:Print((L["Faction %s not found"]):format(name))
	else
		self:Print(name .. ' rep: ' .. self.StandingText[standing])
	end
end

function W:PrintHonor()
	if self.db.profile.Settings.ShowHonorChanges then
		self:Print((L["Honor: %d"]):format(self.Data.Honor))
	end
end

function W:PrintItems()
	if self.db.profile.Settings.ShowItemsChanges then
		local index, value
		for index, value in pairs(self.Data.Items) do
			self:Print((L["Items"]):format(index, value))
		end
		for index, value in pairs(self.db.profile.RequirementItems) do
			self:Print((L["Items"]):format(index, value))
		end
	end
end

function W:PrintMoney()
	if self.db.profile.Settings.ShowMoneyChanges then
		self:Print((L["Money: %s"]):format(self:GetMoneyText()))
	end
end

function W:RemoveItem(items, index, hideMessage)
	local _, link = GetItemInfo(items[index].ItemId)
	table.remove(items, index)

	self:UpdateTotals()
	self:Update()
	self:UpdateItemsMenu()

	if not hideMessage then
		self:Print((L['Removed %s from watch list']):format(link or L['Not Found']))
	end
end

function W:RemoveRequirementItem(item)
	local reqItem = self.db.profile.RequirementItems[item]

	if reqItem then
		local _, link = GetItemInfo(reqItem.ItemId)
		self.db.profile.RequirementItems[item] = nil

		self:UpdateRequirementsMenu()
		self:UpdateItemsMenu()

		self:Print((L['Removed %s from requirements list']):format(link or L['Not Found']))
	end
end

function W:RemoveRequirement(item, reqIndex)
	table.remove(item.Requirements, reqIndex)

	self:UpdateTotals()
	self:Update()
	self:UpdateItemsMenu()
end

function W:RemoveSet(index)
	local set = self.db.char.Sets[index]
	
	-- Move all items back to root
	local _, item
	for _, item in ipairs(set.Items) do
		table.insert(self.db.char.Items, item)
	end
	
	-- Remove set
	table.remove(self.db.char.Sets, index)
	
	self:UpdateItemsMenu()
end

function W:RequirementToNeeded(req)
	if req then
		if req.Type == 'Honor' then
			return (L['%d honor']):format(req.Honor - self.Data.Honor)
		elseif req.Type == 'Money' then
			return self:FormatMoney(req.Money - self.Data.Money, true)
		elseif req.Type == 'Arena' then
			return (L['%d Arena Point(s)']):format(req.ArenaPoints - self.Data.ArenaPoints)
		elseif req.Type == 'Faction' then
			return (L['%s|r with %s']):format(self.StandingText[req.Standing], req.Faction)
		elseif req.Type == 'Item' then
			local count = self:GetItemCount(req.Name)
			return (L['%dx %s']):format(req.Required - count, self:GetShortName(req.Name))
		elseif req.Type == 'Note' then
			return req.Note
		end
	end
	return ''
end

function W:RequirementToString(req)
	if req then
		if req.Type == 'Honor' then
			return (L['%d honor']):format(req.Honor), self.Textures.Honor
		elseif req.Type == 'Money' then
			return self:FormatMoney(req.Money, true), self.Textures.Money
		elseif req.Type == 'Arena' then
			return (L['%d Arena Point(s)']):format(req.ArenaPoints), self.Textures.ArenaPoints
		elseif req.Type == 'Faction' then
			return (L['%s|r with %s']):format(self.StandingText[req.Standing], req.Faction)
		elseif req.Type == 'Item' then
			return (L['%dx %s']):format(req.Required, self:GetShortName(req.Name)), self:GetTexture(req.Name)
		elseif req.Type == 'Note' then
			return req.Note
		end
	end
	return ''
end

function W:RequirementToTooltip(req)
	if req then
		local showNeeded = self.db.profile.Settings.ShowNeeded
		if req.Type == 'Honor' then
			return L['Honor'], ('|cffffffff%d/%d'):format(showNeeded and req.Honor - self.Data.Honor or self.Data.Honor, req.Honor), self.Data.Honor / req.Honor
		elseif req.Type == 'Money' then
			return L['Money'], self:FormatMoney(showNeeded and req.Money - self.Data.Money or self.Data.Honor, true), self.Data.Money / req.Money
		elseif req.Type == 'Arena' then
			return L['Arena Points'], ('|cffffffff%d/%d'):format(showNeeded and req.ArenaPoints - self.Data.ArenaPoints or self.Data.ArenaPoints, req.ArenaPoints), self.Data.ArenaPoints / req.ArenaPoints
		elseif req.Type == 'Faction' then
			local faction = self:GetFaction(req.Faction)
			local earnedValue = faction.EarnedValue
			local needed = self:StandingToValue(req.Standing)
			local standing = showNeeded and req.Standing or faction.Standing
			return req.Faction, (L['%s|r |cffffffff(%d)']):format(self.StandingText[standing], needed - earnedValue), earnedValue / needed
		elseif req.Type == 'Item' then
			local count = self:GetItemCount(req.Name)
			return self:GetShortName(req.Name), ('|cffffffff%d/%d'):format(showNeeded and req.Required - count or count, req.Required), count / req.Required
		elseif req.Type == 'Note' then
			return L['Note'], '|cffffffff' .. req.Note, req.Done and 1 or 0
		end
	end
	return '', '', 0
end

function W:StandingTextToValue(text)
  local s
  for s = 1, 8 do
    if text == self.StandingText[s] then
      return s
    end
  end
  return 0
end

function W:StandingToValue(standing)
  if standing <= 5 then
    return 3000
  elseif standing == 6 then
    return 9000
  elseif standing == 7 then
    return 21000
  else
    return 42000
  end
end

function W:UpdateItems()
	-- Check each item
	local isChanged = false
	local item, value
	for item, value in pairs(self.Data.Items) do
		local count = GetItemCount(item, true)
		if value ~= count then
			self.Data.Items[item] = count
			if self.db.profile.Settings.ShowItemsChanges then
				self:Print((L["Items"]):format(item, count))
			end
			isChanged = true
		end
	end
	for item, reqItem in pairs(self.db.profile.RequirementItems) do
		if reqItem then
			local count = GetItemCount(item, true)
			if reqItem.Count ~= count then
				self.db.profile.RequirementItems[item].Count = count
				if self.db.profile.Settings.ShowItemsChanges then
					self:Print((L["Items"]):format(item, count))
				end
				isChanged = true
			end
		end
	end

	-- Check wished items
	isChanged = isChanged or self:CheckForItems(self.db.char.Items)
	
	-- Check from sets
	local index, set
	for index, set in ipairs(self.db.char.Sets) do
		isChanged = isChanged or self:CheckForItems(set.Items)
	end

	return isChanged
end

function W:UpdateTotals()
	-- Reset totals
	self.Totals = { Requirements = {} }

	-- Combine them
	self:UpdateTotalsFromGroup(self.db.char.Items)
	
	-- Combine from different sets
	local index, set
	for index, set in ipairs(self.db.char.Sets) do
		self:UpdateTotalsFromGroup(set.Items)
	end
end

function W:UpdateTotalsFromGroup(items)
	local index, item
	for index, item in ipairs(items) do
		self:CombineRequirements(self.Totals.Requirements, item.Requirements)
	end
end

----------------------
--- Event Handlers ---
----------------------

function W:BAG_UPDATE()
	if self:UpdateItems() then
		self:Update()
	end
end

function W:PLAYER_MONEY()
	-- Keep money
	self.Data.Money = GetMoney()
	self:PrintMoney()
	self:Update()
end

function W:UPDATE_FACTION()
	-- Keep faction reputation
	for factionIndex = 1, GetNumFactions() do
		local name, _, standingId, _, _, earnedValue, _, _, isHeader = GetFactionInfo(factionIndex)
		if not isHeader then
			self.Data.Factions[name] = { Standing = standingId, EarnedValue = earnedValue }
		end
	end
	self:Update()
end

-----------------------
--- Options Methods ---
-----------------------

W.ItemsMenu = {}
W.RequirementItemsMenu = {}
W.Options = {
    name = 'Wishlist',
    handler = Wishlist,
	type = 'group',
	args = {
		Items = { type = 'group', name = L['Options Items'], desc = L['Options Items'], order = 1, args = W.ItemsMenu },
		Settings = { -- TODO: Inline getter functions
			type = 'group', name = L['Options Settings'], desc = L['Options Settings'], order = 2, args = {
				Text = {
					type = 'group', name = L['Options Text'], desc = L['Options Text'], args = {
						ShowMoney = { type = 'toggle', name = L['Show Money'], desc = L['Show Money'], get = 'IsShowingMoney', set = 'ToggleShowMoney' },
					},
				},
				ShowMoneyChanges = { type = 'toggle', name = L['Show Money Changes'], desc = L['Show Money Changes'], get = 'IsShowingMoneyChanges', set = 'ToggleShowMoneyChanges' },
				ShowHonorChanges = { type = 'toggle', name = L['Show Honor Changes'], desc = L['Show Honor Changes'], get = 'IsShowingHonorChanges', set = 'ToggleShowHonorChanges' },
				ShowItemsChanges = { type = 'toggle', name = L['Show Items Changes'], desc = L['Show Items Changes'], get = 'IsShowingItemsChanges', set = 'ToggleShowItemsChanges' },
				ShowNeeded = { type = 'toggle', name = L['Show Needed'], desc = L['Show Needed'], get = 'IsShowingNeeded', set = 'ToggleShowNeeded' },
				ShowAll = { type = 'toggle', name = L['Show All'], desc = L['Show All'], get = 'IsShowingAll', set = 'ToggleShowAll' },
				AddSet = { type = 'input', name = L['Add Set'], desc = L['Add Set'], usage = L['Add Set'], get = false, set = function (info, v) self:AddSet(v) end },
			},
		},
		Add = { type = 'input', name = L['Add Item'], desc = L['Add Item'], usage = L['Add Item'], get = false, set = function (info, v) self:AddItem(v) end, order = 3 },
		Requirements = { type = 'group', name = L['Requirements'], desc = L['Requirements'], order = 4, args = {
				Items = { type = 'group', name = L['Options Items'], desc = L['Options Items'], args = W.RequirementItemsMenu },
			},	
		},
	}
}

function W:RegisterMenu()
	self:UpdateItemsMenu()
	self:UpdateRequirementsMenu()

	if self.db.profile.Settings.UseAtlasLoot then
		WishlistAtlasLoot:Hook()
	end

    --self.Options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
    LibStub('AceConfig-3.0'):RegisterOptionsTable('Wishlist', self.Options, {'/wishlist', '/wl'});
	self:RegisterChatCommand('wishlist', self.Options)
	self:RegisterChatCommand('wl', self.Options)
	self.OnMenuRequest = self.Options
    
    self.OptionsFrame = acd:AddToBlizOptions("Wishlist")
end

function W:BuildCustom(item)
	local items = {}
	local name, _
	for name, _ in pairs(self.Data.Items) do
		local shortName = self:GetShortName(name)
		table.insert(items, {
			type = 'input',
			name = shortName,
			desc = name,
			usage = name,
			icon = self.Textures[name],
			get = false,
			set = function(info, v) self:AddRequirements(item, { { Type = 'Item', Name = name, Required = tonumber(v) } }) end,
		})
	end
	for name, reqItem in pairs(self.db.profile.RequirementItems) do
		if reqItem then
			local _, _, _, _, _, _, _, _, _, itemTexture = GetItemInfo(reqItem.ItemId)
			table.insert(items, {
				type = 'input',
				name = name,
				desc = name,
				usage = name,
				icon = itemTexture,
				get = false,
				set = function(info, v) self:AddRequirements(item, { { Type = 'Item', Name = name, Required = tonumber(v) } }) end,
			})
		end
	end
	-- TODO: Sort all items
	table.sort(items, function (a, b) return a.name < b.name end)

	local factions = {}
	local factionValidate = { [0] = self.StandingText[5], [1] = self.StandingText[6], [2] = self.StandingText[7], [3] = self.StandingText[8] }
	for index, name in ipairs(self.Factions) do
		table.insert(factions, {
			type = 'input',
			name = name,
			desc = name,
			usage = name,
			validate = factionValidate,
			order = index,
			get = false,
			set = function(info, v) self:AddRequirements(item, { { Type = 'Faction', Faction = name, Standing = self:StandingTextToValue(v) } }) end,
		})
	end

	return {
		{
			type = 'input',
			name = L['Honor'],
			desc = L['Honor'],
			usage = L['Honor'],
			order = 1,
			icon = self.Textures.Honor,
			get = false,
			set = function(info, v) self:AddRequirements(item, { { Type = 'Honor', Honor = tonumber(v) } }) end,
		},
		{
			type = 'input',
			name = L['Arena Points'],
			desc = L['Arena Points'],
			usage = L['Arena Points'],
			order = 2,
			icon = self.Textures.ArenaPoints,
			get = false,
			set = function(info, v) self:AddRequirements(item, { { Type = 'Arena', ArenaPoints = tonumber(v) } }) end,
		},
		{
			type = 'input',
			name = L['Money (copper)'],
			desc = L['Money (copper)'],
			usage = L['Money (copper)'],
			order = 3,
			icon = self.Textures.Money,
			get = false,
			set = function(info, v) self:AddRequirements(item, { { Type = 'Money', Money = tonumber(v) } }) end,
		},
		{
			type = 'group',
			name = L['Options Items'],
			desc = L['Options Items'],
			order = 4,
			args = items,
		},
		{
			type = 'group',
			name = L['Factions'],
			desc = L['Factions'],
			order = 5,
			args = factions,
		},
		{
			type = 'input',
			name = L['Note'],
			desc = L['Note'],
			usage = L['Note'],
			order = 6,
			get = false,
			set = function(info, v) self:AddRequirements(item, { { Type = 'Note', Note = v } }) end,
		},
	}
end

function W:BuildPredefined(item, list)
	list = list or data.Predefined

	local result = {}
	local index, predefined
	for index, predefined in ipairs(list) do
		local args = predefined.Args
		if args then
			table.insert(result, { type = 'group', name = predefined.Name, desc = predefined.Name, order = index, icon = predefined.Icon, args = self:BuildPredefined(item, args) })
		else
			table.insert(result, { type = 'execute', name = predefined.Name, desc = predefined.Name, order = index, icon = predefined.Icon, func = function() self:AddRequirements(item, predefined.Requirements) end })
		end
	end

	return result
end

function W:UpdateAddToItemsMenu(items, menu, set)
	local index, item
	for index, item in ipairs(items) do
		local itemName, _, _, _, _, _, _, _, _, itemTexture = GetItemInfo(item.ItemId)
		itemName = itemName or L['Not Found']
		
		local itemMenu = {}
		local reqIndex, req
		for reqIndex, req in ipairs(item.Requirements) do
			local reqString, icon = self:RequirementToString(req)
			table.insert(itemMenu, {
				type = 'group',
				name = reqString,
				desc = reqString,
				icon = icon,
				order = reqIndex,
				args = {
					RemoveRequirement = {
						type = 'execute',
						name = L['Remove'],
						desc = L['Remove'],
						order = 2,
						func = function() self:RemoveRequirement(item, reqIndex) end,
					},
				},
			})
		end
		
		table.insert(itemMenu, {
			type = 'group',
			name = L['Add Custom'],
			desc = L['Add Custom'],
			order = 997,
			args = self:BuildCustom(item),
		})
		table.insert(itemMenu, {
			type = 'group',
			name = L['Add Predefined'],
			desc = L['Add Predefined'],
			order = 998,
			args = self:BuildPredefined(item),
		})
		table.insert(itemMenu, {
			type = 'execute',
			name = L['Remove Item'],
			desc = L['Remove Item'],
			order = 999,
			func = function() self:RemoveItem(items, index) end,
		})
		if set and #self.db.char.Sets > 1 or #self.db.char.Sets > 0 then
      local moveOptions = {}
 	    local setIndex, otherSet
	    for setIndex, otherSet in ipairs(self.db.char.Sets) do
	      if otherSet ~= set then
	        table.insert(moveOptions, {
	          type = 'execute',
	          name = otherSet.Name,
	          desc = otherSet.Name,
	          order = setIndex,
	          func = function() self:MoveItem(items, index, otherSet.Items) end,
	        })
	      end
	    end
	    if set then
	        table.insert(moveOptions, {
	          type = 'execute',
	          name = (L['Remove from %s']):format(set.Name),
	          desc = (L['Remove from %s']):format(set.Name),
	          order = 999,
	          func = function() self:MoveItem(items, index, self.db.char.Items) end,
	        })
	    end
     table.insert(itemMenu, { type = 'group', name = L['Move to Set'], desc = L['Move to Set'], order = 1000, args = moveOptions })
		end
		table.insert(menu, { type = 'group', name = itemName, desc = itemName, order = index, icon = itemTexture, args = itemMenu })
	end
end

function W:UpdateItemsMenu()
	-- Remove everything
	while #self.ItemsMenu > 0 do table.remove(self.ItemsMenu) end
	self.ItemsMenu = {} -- TEMP: Don't know how to remove Set menu items

	-- Add items not in set
	self:UpdateAddToItemsMenu(self.db.char.Items, self.ItemsMenu)
		
	-- Add sets with items
	local itemsCount = #self.db.char.Items
	local index, set
	for index, set in ipairs(self.db.char.Sets) do
	  local setMenu = {}
	  self:UpdateAddToItemsMenu(set.Items, setMenu, set)
	  
	  -- Add commands
	  setMenu.Add = { type = 'input', name = L['Add Item'], desc = L['Add Item'], usage = L['Add Item'], get = false, set = function(info, v) self:AddItem(v, set.Items) end, order = 999 }
	  setMenu.Remove = { type = 'execute', name = L['Remove Set'], desc = L['Remove Set'], func = function() self:RemoveSet(index) end, order = 1000 }
	  
	  -- Add to menu
	  self.ItemsMenu[set.Name] = { type = 'group', name = set.Name, desc = set.Name, order = index + itemsCount, args = setMenu }
	end

	self.Options.args.Items.args = self.ItemsMenu
end

function W:UpdateRequirementsMenu()
	-- Update requirements menu
	while #self.RequirementItemsMenu > 0 do table.remove(self.RequirementItemsMenu) end
	
	local item
	for index, item in pairs(self.db.profile.RequirementItems) do
		if item then
			local itemName, _, _, _, _, _, _, _, _, itemTexture = GetItemInfo(item.ItemId)
			itemName = itemName or L['Not Found']

			table.insert(self.RequirementItemsMenu, { type = 'group', name = itemName, desc = itemName, icon = itemTexture, args = { { type = 'execute', name = L['Remove Item'], desc = L['Remove Item'], func = function() self:RemoveRequirementItem(index) end } } })
		end
	end
	self.RequirementItemsMenu.Add = { type = 'input', name = L['Add Item'], desc = L['Add Item'], usage = L['Add Item'], get = false, set = function(info, v) self:AddRequirementItem(v) end, order = 999 }

	--self.Options.args.Requirements.args.Items.args = self.RequirementItemsMenu
end

function W:IsShowingAll()
	return self.db.profile.Settings.ShowAll
end
function W:IsShowingHonorChanges()
	return self.db.profile.Settings.ShowHonorChanges
end

function W:IsShowingItemsChanges()
	return self.db.profile.Settings.ShowItemsChanges
end

function W:IsShowingMoney()
	return self.db.profile.Text.ShowMoney
end

function W:IsShowingMoneyChanges()
	return self.db.profile.Settings.ShowMoneyChanges
end

function W:IsShowingNeeded()
	return self.db.profile.Settings.ShowNeeded
end

function W:ToggleShowAll()
	self.db.profile.Settings.ShowAll = not self.db.profile.Settings.ShowAll
	self:OnDataUpdate()
	return self.db.profile.Settings.ShowAll
end

function W:ToggleShowHonorChanges()
	self.db.profile.Settings.ShowHonorChanges = not self.db.profile.Settings.ShowHonorChanges
	self:PrintHonor()
	return self.db.profile.Settings.ShowHonorChanges
end

function W:ToggleShowItemsChanges()
	self.db.profile.Settings.ShowItemsChanges = not self.db.profile.Settings.ShowItemsChanges
	self:PrintItems()
	return self.db.profile.Settings.ShowItemsChanges
end
function W:ToggleShowMoney()
	self.db.profile.Text.ShowMoney = not self.db.profile.Text.ShowMoney
	self:OnTextUpdate()
	return self.db.profile.Text.ShowMoney
end

function W:ToggleShowMoneyChanges()
	self.db.profile.Settings.ShowMoneyChanges = not self.db.profile.Settings.ShowMoneyChanges
	self:PrintMoney()
	return self.db.profile.Settings.ShowMoneyChanges
end

function W:ToggleShowNeeded()
	self.db.profile.Settings.ShowNeeded = not self.db.profile.Settings.ShowNeeded
	return self.db.profile.Settings.ShowNeeded
end